/*
 *  linux/drivers/video/mach64.c -- Low level frame buffer operations for 
 *                                  Ati Mach64 (accelerated)
 *
 *	Created 20 Jun 1997 by Bernd Harries
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>
#include <linux/config.h>
#include <linux/fb.h>

#include "fbcon.h"
#include "fbcon-mach64.h"

/*============================================================================*/
/* Defines */

#define  _USE_HW_ACCEL_BLT_
#define  _USE_HW_ACCEL_FILL_


#ifdef  BHA_DEBUG

#define  FuncId       if(mach64_debug & 0x00000100) printk
#define  FuncDone     if(mach64_debug & 0x00000200) printk
#define  CFuncId      if(mach64_debug & 0x00000400) printk
#define  CFuncDone    if(mach64_debug & 0x00000800) printk


/* Busy */
#define  Bprint       if(mach64_debug & 0x01000000) printk
/* Debug */
#define  Dprint       if(mach64_debug & 0x00800000) printk

/* Message */
#define  Mprint       if(mach64_debug & 0x00004000) printk
/* Info */
#define  Iprint       if(mach64_debug & 0x00008000) printk
/* Warn */
#define  Wprint       if(mach64_debug & 0x00010000) printk
/* Error */
#define  Eprint       if(mach64_debug & 0x00020000) printk
/* Fatal */
#define  Fprint       if(mach64_debug & 0x00040000) printk
/* Panic */
#define  Pprint       if(mach64_debug & 0x00040000) printk

#else

#define  FuncId
#define  FuncDone
#define  CFuncId
#define  CFuncDone

#endif  /* BHA_DEBUG*/

/*============================================================================*/
/* Types */

/*============================================================================*/
/* Extern Functions*/



/*============================================================================*/
/* Intern Functions*/

    /*
     *  Prototypes
     */

static int open_mach64(struct display *p);
static void release_mach64(void);
static void bmove_mach64(struct display *p, int sy, int sx, int dy, int dx,
		       int height, int width);
static void clear_mach64(struct vc_data *conp, struct display *p, int sy,
		       int sx, int height, int width);
static void putc_mach64(struct vc_data *conp, struct display *p, int c, int yy,
		      int xx);
static void putcs_mach64(struct vc_data *conp, struct display *p,
		       const char *s, int count, int yy, int xx);
static void rev_char_mach64(struct display *p, int xx, int yy);


/*============================================================================*/
/*============================================================================*/
/* Intern Variables */

/*--------------------------------------------------------------------------*/
/* BSS */

/*--------------------------------------------------------------------------*/
/* Data */

    /*
     *  `switch' for the low level operations
     */

static struct display_switch    dispsw_mach64 = 
{
  open_mach64,
  release_mach64,
  bmove_mach64,
  clear_mach64,
  putc_mach64,
  putcs_mach64,
  rev_char_mach64
};


    /*
     *  8 bpp packed pixels => 4 pixel / 32 Bit => 1/2 FontByte 
     */

static u_int nibbletab_mach64[16] = 
{
  0x00000000, 0x000000ff, 0x0000ff00, 0x0000ffff,
  0x00ff0000, 0x00ff00ff, 0x00ffff00, 0x00ffffff,
  0xff000000, 0xff0000ff, 0xff00ff00, 0xff00ffff,
  0xffff0000, 0xffff00ff, 0xffffff00, 0xffffffff
};

static  unsigned int    mach64_debug = 0x018FCF00;

/*============================================================================*/
/*============================================================================*/
/* Functions */

/*============================================================================*/
/*============================================================================*/


/*--------------------------------------------------------------------------*/

static int 
open_mach64(struct display * p)

/*--------------------------------------------------------------------------*/
{
  /*--------------------------------------------------------------------------*/
  FuncId("open_mach64($%08lX) : ", p);
  if(p->type != FB_TYPE_PACKED_PIXELS || p->var.accel != FB_ACCEL_MACH64)
  {
    FuncDone(" Fail! \n");
    return -EINVAL;  /* This driver can only do Mach64 boards! */
  }
  /*endif*/

  p->next_line = (p->var.xres_virtual * p->var.bits_per_pixel) >> 3;
  p->next_plane = 0;
  MOD_INC_USE_COUNT;

  FuncDone(" Match! \n");
  return 0;
}
/*endproc open_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
release_mach64(void)
{
  /*--------------------------------------------------------------------------*/
  FuncId("release_mach64() \n");
    MOD_DEC_USE_COUNT;
}
/*endproc release_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
bmove_mach64(struct display * p,
  int sy,
  int sx,
  int dy,
  int dx,
  int height,
  int width)

/*--------------------------------------------------------------------------*/
{
  u_char * src, * dst;
  int                           bytes = p->next_line,
                                linesize = bytes * p->fontheight,
                                rows;

  /*--------------------------------------------------------------------------*/
  CFuncId("bmove_mach64($%04X, $%04X, $%04X, $%04X) \n",
    sy, sx, dy, dx, height, width);

#ifdef _USE_HW_ACCEL_BLT_

  Mach64_BitBLT(
    (sx << 3),
    (sy * p->fontheight),
    (dx << 3),
    (dy * p->fontheight),
    (width << 3),
    (height * p->fontheight),
    3);                 /*  MIX_SRC); see XGCValues GCFunction = GXCopy */

#else

  if(sx == 0 && dx == 0 && width * 8 == bytes)   /* full stride */

    mymemmove(p->screen_base + dy * linesize,
      p->screen_base + sy * linesize,
      height * linesize);

  else if (dy < sy || (dy == sy && dx < sx))
  {
    src = p->screen_base + sy * linesize + sx * 8;
    dst = p->screen_base + dy * linesize + dx * 8;
    for (rows = height * p->fontheight ; rows-- ;)
    {
      mymemmove(dst, src, (width << 3));
      src += bytes;
      dst += bytes;
    }
    /*next rows*/
  }
  else
  {
    src = p->screen_base + (sy+height) * linesize + sx * 8 - bytes;
    dst = p->screen_base + (dy+height) * linesize + dx * 8 - bytes;
    for(rows = height * p->fontheight ; rows-- ;)
    {
      mymemmove(dst, src, (width << 3));
      src -= bytes;
      dst -= bytes;
    }
    /*next rows*/
  }
  /*endif*/

#endif  /* _USE_HW_ACCEL_BLT_ */

  return;
}
/*endproc bmove_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
clear_mach64(struct vc_data * conp,
  struct display * p,
  int sy,                       /* in Characters von oben */
  int sx,
  int height,                   /* in Characters */
  int width)                    /* in Characters */

/*--------------------------------------------------------------------------*/
{
  register u_int *              dest;
  register u_int                bgx;
  register int                  i;

  u_char *                      dest0;

  int                           bytes_p_line = p->next_line,
                                lines_w = height * p->fontheight,
                                rows;

  /*--------------------------------------------------------------------------*/
  CFuncId("clear_mach64($%04X, $%04X, $%04X, $%04X) \n",
    sy, sx, height, width);

  bgx = attr_bgcol_ec(p, conp);
  bgx |= (bgx << 8);
  bgx |= (bgx << 16);

#ifdef _USE_HW_ACCEL_FILL_

  Mach64_RectFill(
    (sx << 3),
    (sy * p->fontheight),
    (width << 3),
    (height * p->fontheight),
    3,                            /* MIX_SRC, see XGCValues GCFunction  */
    bgx);

#else

  dest0 = p->screen_base + sy * p->fontheight * bytes_p_line + sx * 8;
  if(sx == 0 && (width << 3) == bytes_p_line)     /* full stride */
  {
    lines_w *= width;
    dest = (u_int*)dest0;
    for(i = 0 ; i < lines_w ; i++)
    {
      *dest++ = bgx;          /* clears 4 pixels == half Char */
      *dest++ = bgx;
    }
  }
  else                                     /* only part of the width */
  {
    for(rows = lines_w; rows-- ; dest0 += bytes_p_line)
    {
      dest = (u_int*)dest0;
      for(i = 0 ; i < width ; i++)
      {
        *dest++ = bgx;        /* clears 4 pixels */
        *dest++ = bgx;
      }
      /*next i*/
    }
  }
  /*endif*/

#endif  /* _USE_HW_ACCEL_FILL_ */

  return;
}
/*endproc clear_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
putc_mach64(struct vc_data * conp,
  struct display * p,
  int c,                   /* ASCII */
  int yy,                  /* Text column */
  int xx)                  /* Text row */

/*--------------------------------------------------------------------------*/
{
  register u_char *             dest_p;
  register u_char *             cdat_p;
  register u_char               cdat;
  register u_int                eorx, fgx, bgx;
  int                           bytes = p->next_line; /*must be multiple of 4*/
  int                           rows;

  /*--------------------------------------------------------------------------*/
  c &= 0xFF;

  dest_p = p->screen_base + yy * p->fontheight * bytes + (xx << 3);
  cdat_p = p->fontdata + c * p->fontheight;

  fgx=attr_fgcol(p, conp);
  bgx=attr_bgcol(p,conp);

  fgx |= (fgx << 8);
  fgx |= (fgx << 16);
  bgx |= (bgx << 8);
  bgx |= (bgx << 16);
  eorx = fgx ^ bgx;

  /* Mach64_WaitBlit(); */    /* We will have to sync with HW Blit */
  for(rows = p->fontheight ; rows-- ; dest_p += bytes) /* byte* add = natural */
  {
    cdat = *cdat_p++;
    *((u_int*)dest_p) = (nibbletab_mach64[cdat >> 4] & eorx) ^ bgx;
    ((u_int *)dest_p)[1] = (nibbletab_mach64[cdat & 0xF] & eorx) ^ bgx;
  }
  /*next row*/

  return;
}
/*endproc putc_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
putcs_mach64(struct vc_data * conp,
  struct display * p,
  const char * s,
  int count,
  int yy,
  int xx)

/*--------------------------------------------------------------------------*/
{
  register u_char *             dest_p;
  register u_char *             cdat_p;
  register u_char               cdat;
  register u_int                eorx, fgx, bgx;
  u_char *                      dest0;
  int                           bytes = p->next_line; /*must be multiple of 4*/
  int                           rows;
  u_char                        c;

  /*--------------------------------------------------------------------------*/

  dest0 = p->screen_base + yy * p->fontheight * bytes + xx * 8;
  fgx=attr_fgcol(p,conp);
  bgx=attr_bgcol(p,conp);

  fgx |= (fgx << 8);
  fgx |= (fgx << 16);
  bgx |= (bgx << 8);
  bgx |= (bgx << 16);
  eorx = fgx ^ bgx;

  /* Mach64_WaitBlit(); */    /* We will have to sync with HW Blit */
  while(count--)        /* character count 8 x h */
  {
    c = *s++;
    cdat_p = p->fontdata + c * p->fontheight;

    for(rows = p->fontheight, dest_p = dest0; rows-- ; dest_p += bytes)
    {
      cdat = *cdat_p++;
      *((u_int*)dest_p) = (nibbletab_mach64[cdat >> 4] & eorx) ^ bgx;
      ((u_int *)dest_p)[1] = (nibbletab_mach64[cdat & 0xF] & eorx) ^ bgx;
    }
    /*next rows*/
    dest0 += 8;
  }
  /*endwhile*/

  return;
}
/*endproc putcs_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
rev_char_mach64(struct display *p, int xx, int yy)

/*--------------------------------------------------------------------------*/
{
    u_char *dest;
    int bytes=p->next_line, rows;

    dest = p->screen_base + yy * p->fontheight * bytes + xx * 8;
    for (rows = p->fontheight ; rows-- ; dest += bytes) {
	((u_long *)dest)[0] ^= 0x0f0f0f0f;
	((u_long *)dest)[1] ^= 0x0f0f0f0f;
    }
}
/*endproc rev_char_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

#ifdef MODULE
int 
init_module(void)
#else
int 
fbcon_init_mach64(void)
#endif

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  FuncId("fbcon_init_mach64() \n");
  /* We want to be added to the list of accelerated drivers */
  return(fbcon_register_driver(&dispsw_mach64, 1));
}
/*endproc fbcon_init_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

#ifdef MODULE
void 
cleanup_module(void)

/*--------------------------------------------------------------------------*/
{
  /*--------------------------------------------------------------------------*/
  FuncId("cleanup_module() \n");
    fbcon_unregister_driver(&dispsw_mach64);
}
/*endproc cleanup_mach64() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
#endif /* MODULE */
/*============================================================================*/


/*============================================================================*/
/* RCS-ID */

  static char
  rcs_id_c[] =
    "@(#) Linux-m68k: 1997 08 16 12:39:32 BHA: fbcon-mach64.c ";

  static char
  rcs_id_h[] = FBCON_MACH64_H_ID;

/*============================================================================*/
